package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.repository.MaterialRepository
import dev.fyloz.colorrecipesexplorer.rest.FILE_CONTROLLER_PATH
import dev.fyloz.colorrecipesexplorer.service.config.ConfigurationService
import io.jsonwebtoken.lang.Assert
import org.springframework.context.annotation.Lazy
import org.springframework.context.annotation.Profile
import org.springframework.stereotype.Service
import java.net.URLEncoder
import java.nio.charset.StandardCharsets

interface MaterialService :
    ExternalNamedModelService<Material, MaterialSaveDto, MaterialUpdateDto, MaterialOutputDto, MaterialRepository> {
    /** Checks if a material with the given [materialType] exists. */
    fun existsByMaterialType(materialType: MaterialType): Boolean

    /** Checks if the given [material] has a SIMDUT file. */
    fun hasSimdut(material: Material): Boolean

    /** Gets all materials that are not a mix type. */
    fun getAllNotMixType(): Collection<MaterialOutputDto>

    /** Gets all materials available for the creation of a mix for the recipe with the given [recipeId], including normal materials and materials from [MixType]s included in the said recipe. */
    fun getAllForMixCreation(recipeId: Long): Collection<MaterialOutputDto>

    /** Gets all materials available for updating the mix with the given [mixId], including normal materials and materials from [MixType]s included in the mix recipe, excluding the material of the [MixType] of the said mix. */
    fun getAllForMixUpdate(mixId: Long): Collection<MaterialOutputDto>

    /** Updates the quantity of the given [material] with the given [factor] and returns the updated quantity. */
    fun updateQuantity(material: Material, factor: Float): Float
}

@Service
@Profile("!emergency")
class MaterialServiceImpl(
    materialRepository: MaterialRepository,
    val recipeService: RecipeService,
    val mixService: MixService,
    @Lazy val materialTypeService: MaterialTypeService,
    val fileService: FileService,
    val configService: ConfigurationService
) :
    AbstractExternalNamedModelService<Material, MaterialSaveDto, MaterialUpdateDto, MaterialOutputDto, MaterialRepository>(
        materialRepository
    ),
    MaterialService {
    override fun idNotFoundException(id: Long) = materialIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = materialIdAlreadyExistsException(id)
    override fun nameNotFoundException(name: String) = materialNameNotFoundException(name)
    override fun nameAlreadyExistsException(name: String) = materialNameAlreadyExistsException(name)

    override fun Material.toOutput(): MaterialOutputDto =
        MaterialOutputDto(
            id = this.id!!,
            name = this.name,
            inventoryQuantity = this.inventoryQuantity,
            isMixType = this.isMixType,
            materialType = this.materialType!!,
            simdutUrl = if (fileService.exists(this.simdutFilePath))
                "${configService.getContent(ConfigurationType.INSTANCE_URL)}$FILE_CONTROLLER_PATH?path=${
                    URLEncoder.encode(
                        this.simdutFilePath,
                        StandardCharsets.UTF_8
                    )
                }"
            else null
        )

    override fun existsByMaterialType(materialType: MaterialType): Boolean =
        repository.existsByMaterialType(materialType)

    override fun hasSimdut(material: Material): Boolean = fileService.exists(material.simdutFilePath)
    override fun getAllNotMixType(): Collection<MaterialOutputDto> = getAllForOutput().filter { !it.isMixType }

    override fun save(entity: MaterialSaveDto): Material =
        save(with(entity) {
            material(
                name = entity.name,
                inventoryQuantity = entity.inventoryQuantity,
                materialType = materialTypeService.getById(materialTypeId),
                isMixType = false
            )
        }).apply {
            if (entity.simdutFile != null && !entity.simdutFile.isEmpty) fileService.write(
                entity.simdutFile,
                this.simdutFilePath,
                false
            )
        }

    override fun update(entity: MaterialUpdateDto): Material {
        val persistedMaterial by lazy {
            getById(entity.id).apply { assertPersistedMaterial(this) }
        }

        return update(with(entity) {
            material(
                id = id,
                name = if (name != null && name.isNotBlank()) name else persistedMaterial.name,
                inventoryQuantity = if (inventoryQuantity != null && inventoryQuantity != Float.MIN_VALUE) inventoryQuantity else persistedMaterial.inventoryQuantity,
                isMixType = persistedMaterial.isMixType,
                materialType = if (materialTypeId != null) materialTypeService.getById(materialTypeId) else persistedMaterial.materialType
            )
        }).apply {
            if (entity.simdutFile != null && !entity.simdutFile.isEmpty) fileService.write(
                entity.simdutFile,
                this.simdutFilePath,
                true
            )
        }
    }

    override fun updateQuantity(material: Material, factor: Float) = with(material) {
        val updatedQuantity = this.inventoryQuantity + factor
        repository.updateInventoryQuantityById(this.id!!, updatedQuantity)
        updatedQuantity
    }

    override fun getAllForMixCreation(recipeId: Long): Collection<MaterialOutputDto> {
        val recipesMixTypes = recipeService.getById(recipeId).mixTypes
        return getAllForOutput()
            .filter { !it.isMixType || recipesMixTypes.any { mixType -> mixType.material.id == it.id } }
    }

    override fun getAllForMixUpdate(mixId: Long): Collection<MaterialOutputDto> {
        val mix = mixService.getById(mixId)
        val recipesMixTypes = mix.recipe.mixTypes
        return getAllForOutput()
            .filter { !it.isMixType || recipesMixTypes.any { mixType -> mixType.material.id == it.id } }
            .filter { it.id != mix.mixType.material.id }
    }

    private fun assertPersistedMaterial(material: Material) {
        Assert.notNull(material.name, "The persisted material with the id ${material.id} has a null name")
    }

    override fun delete(entity: Material) {
        if (!repository.canBeDeleted(entity.id!!)) throw cannotDeleteMaterialException(entity)
        if (fileService.exists(entity.simdutFilePath)) fileService.delete(entity.simdutFilePath)
        super.delete(entity)
    }
}
